-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module provides a pretty printer for the toolkit



module Common.PrettyPrinter where

-- import Typing
import Common.Absglobaltype
-- import Common.Typing
import Data.Map as M
import Data.List as L


listPrinter :: PrettyPrinter a => String -> String -> [a] -> String
listPrinter _ end [] = end
listPrinter _ end [x] = (prPrint x) ++ end
listPrinter sep _ (x:xs) = (prPrint x) ++ sep ++ (prPrintList xs) 

listComma ::  PrettyPrinter a => [a] -> String
listComma xs = listPrinter (", ") ("") xs

class PrettyPrinter a where
    prPrint :: a -> String
    prPrintList :: [a] -> String


-- instance PrettyPrinter Primitive where
--     prPrint _ = 
--     prPrintList _ = 

-- instance PrettyPrinter Participant where
--     prPrint _ = 
--     prPrintList _ = 

instance PrettyPrinter LocalType where
    prPrint (Send ch ts ass lt) = (prPrint ch)++"!<"++(prPrintList ts)++">["++(prPrint ass)++"];"++(prPrint lt)
    prPrint (LDelegS ch del ass lt) = (prPrint ch)++"!<<"++(prPrint del)++">>["++(prPrint ass)++"];"++(prPrint lt)
    prPrint (Receive ch ts ass lt) =  (prPrint ch)++"?<"++(prPrintList ts)++">["++(prPrint ass)++"];"++(prPrint lt)
    prPrint (LDelegR ch del ass lt) = (prPrint ch)++"?<<"++(prPrint del)++">>["++(prPrint ass)++"];"++(prPrint lt)
    prPrint (Branch ch id lbs) = (prPrint ch)++"&{\n"++(prPrintList lbs)++"\n}"
    prPrint (Select ch id lbs) = (prPrint ch)++"${\n"++(prPrintList lbs)++"\n}"
    prPrint (LRecur v es vd a lt) = "mu "++(prPrint v)++"("++(prPrintList es)++")"++
                                    "{"++(prPrintList vd)++"}["++(prPrint a)++"]."++(prPrint lt)
    prPrint (LRecCall v es) = (prPrint v) ++"("++(prPrintList es)++")"
    prPrint LEnd = "end"
    prPrint Bottom = "_|_"

    prPrintList xs = listPrinter "\n" "" xs


instance PrettyPrinter DelegDecl where
    prPrint (SessPass chs lt p) = "["++(prPrintList chs)++"]: {"++(prPrint lt)++"}@"++(prPrint p)
    prPrintList xs = listPrinter "\n" ";" xs 

instance PrettyPrinter Typ where
    prPrint TInt = "int"
    prPrint TBool = "bool"
    prPrint TString = "string"
    prPrint TDate = "date"

    prPrintList xs = listComma xs


instance PrettyPrinter LabelElmt where
    prPrint (LabelElmt ass l lt) = "["++(prPrint ass)++"]"++(prPrint l)++": "++(prPrint lt)
    prPrintList xs = listPrinter ",\n" "" xs

instance PrettyPrinter Assertion where
    prPrint (Assert b) = prPrint b
    prPrint (AssertEmpty) = ""
    prPrintList xs = listComma xs

instance PrettyPrinter Ident where
    prPrint (Ident s) = s
    prPrintList xs = listComma xs
                        
instance PrettyPrinter Channel where
    prPrint (Chan (Ident s)) = s
    prPrintList xs = listComma xs

instance PrettyPrinter Participant where
    prPrint (Parti (Ident s)) = s
    prPrintList xs = listComma xs


instance PrettyPrinter Variable where
    prPrint (Var (Ident s)) = s
    prPrintList xs = listComma xs

instance PrettyPrinter VarDecl where
    prPrint (VarDeclaration v t) = (prPrint v)++": "++(prPrint t)
    prPrintList xs = listComma xs

instance PrettyPrinter Label where
    prPrint (Lab (Ident s)) = s
    prPrintList xs = listComma xs

instance PrettyPrinter Exp where
    prPrint (ExpI e) = prPrint e
    prPrint (ExpV e) = prPrint e
    prPrint (ExpS e) = e
    prPrint (BTrue) = "true"
    prPrint (BFalse) = "false"
    prPrint (ExpD i1 i2 i3) = (prPrint i1)++"/"++(prPrint i2)++"/"++(prPrint i3)
    prPrint (BNot exp) = "!("++(prPrint exp)++")"
    prPrint (EBin e1 b e2) = (prPrint e1)++" "++(prPrint b)++" "++(prPrint e2)
    prPrint (EAr e1 b e2) = (prPrint e1)++" "++(prPrint b)++" "++(prPrint e2)
    prPrint (EExist vs e) = "Exist "++(prPrintList vs)++" st. ("++(prPrint e)++")"
    prPrint (EForall vs e) = "Forall "++(prPrintList vs)++" st. ("++(prPrint e)++")"

    prPrintList xs = listComma xs

instance PrettyPrinter BinOp where
    prPrint EOr = "||"
    prPrint EAnd = "&&"
    prPrint EEq = "="
    prPrint ENEq = "!="
    prPrint EGt = ">"
    prPrint ELt = "<"
    prPrint ELte = "<="
    prPrint EGte = ">="
    prPrint EImpl = "=>"
    prPrintList xs = listComma xs

instance PrettyPrinter ArOp where
    prPrint EPlus = "+"
    prPrint EDiv = "/"
    prPrint EMul = "*"
    prPrint EMinus = "-"
    prPrintList xs = listComma xs

instance PrettyPrinter Integer where
    prPrint i = (show i)
    prPrintList xs = listComma xs


instance PrettyPrinter DistributedInteraction where
    prPrint x = "delegation not yet supported"
    prPrintList xs = "delegation not yet supported"

instance PrettyPrinter GlobalSpecification where
    prPrint x = show x
    prPrintList xs = listComma xs


instance PrettyPrinter Projection where
    prPrint (Proj p lt) = "["++(prPrint lt)++"]@"++(prPrint p)
    prPrintList xs = listPrinter "\n\n" "" xs



instance PrettyPrinter ProjMap where
    prPrint (ProjMap m) = pr (M.toList m)
        where pr [] = ""
              pr ((c,p):xs) = "*"++(prPrint c)++":\n"
                               ++(prPrintList p)++"\n"
                               ++ pr xs
    prPrintList xs = listPrinter "\n" "" xs



instance PrettyPrinter Sorting where
    prPrint Empty = "[]"
    prPrint (Sorting nm pm prm) = "Variables: "++(prPrint nm) ++"\nProcesses: "++ (prPrint pm) ++"\nProjections: "++ (prPrint prm)
    prPrintList xs = listPrinter "\n" "" xs

instance PrettyPrinter NameMap where
    prPrint (NameMap m) = prPrintList (L.map f (M.toList m))
                        where f (v,t) = VarDeclaration v t 
    prPrintList xs = listPrinter "\n" "" xs

instance PrettyPrinter ProcessMap where
    prPrint (ProcessMap m) = show m
    prPrintList xs = listPrinter "\n" "" xs
