-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module gathers a set of various functions and data types
-- used throughout the program
--


module Common.Utils where

import Common.Absglobaltype
import Data.Set as Set
import Data.List as List
import Data.Map as M
import Common.PrettyPrinter


--
-- Represent a Global types prefix 
-- (i.e. used to test linearity)
--
data Prefix = Prefix Ident Ident Ident deriving (Eq,Ord,Show)


--
-- Substitution to replace variables by expressions
--
type Substitution = ([Exp], [Variable])

emptySub :: Substitution
emptySub = ([],[])

{- ------------------------------------------------------------------- -}
{-                           GENERAL                                   -}
{- ------------------------------------------------------------------- -}
-- 
-- Returns true if intersection of all sets is emtpy
--
intersects :: (Ord a) => [Set a] -> Bool
intersects [] = True
intersects (x:[]) = True
intersects (x:y:xs) = Set.null (Set.intersection x y) && (intersects (x:xs)) && (intersects (y:xs))



{- ------------------------------------------------------------------- -}
{-                           CHANNELS                                  -}
{- ------------------------------------------------------------------- -}

--
-- Returns the set of channels used in a distributed interactions
--
channelList :: DistributedInteraction -> Set Channel
channelList (ParallelInter []) = Set.empty
channelList (ParallelInter (Conc x:xs)) = 
    case x of
      SingleInter y -> Set.union (channelList (SingleInter y))
                       (channelList (ParallelInter xs))
      ParallelInter ((Conc y):ys) -> Set.union (channelList y) 
                                     (Set.union (channelList (ParallelInter ys))
                                      (channelList (ParallelInter xs)))
channelList (SingleInter x) = case x of
                                ValPass _ _ c _ _ int -> Set.insert c (channelList int)
                                Deleg _ _ c _ _ int -> Set.insert c (channelList int)
                                LabelPass _ _ c _ br -> channelList_B br
                                otherwise -> Set.empty
-- 
-- channelList helper (for Select/Branch nodes)
--
channelList_B :: [BranchElmt] -> Set Channel
channelList_B [] = Set.empty
channelList_B (x:xs) = case x of
                         BranchingElmt _ _ int -> Set.union (channelList int) (channelList_B xs)




{- ------------------------------------------------------------------- -}
{-                      DATA TYPE TRAVERSING                           -}
{- ------------------------------------------------------------------- -}

--
-- Traverse Process Data Type
--
getChildProcess :: Primitive -> Process
getChildProcess (PInit _ _ _ pr) = pr
getChildProcess (PJoin _ _ _ pr) = pr
getChildProcess (PSend _ _ _ _ pr) = pr
getChildProcess (PDelegS _ _ _ _ pr) = pr
getChildProcess (PReceive _ _ _ pr) = pr
getChildProcess (PDelegR _ _ _ pr) = pr
getChildProcess (PSelect _ _ _ _ pr) = pr
getChildProcess l@(PBranch _ _ _) = error ("Cannot get one child of a branch node: "++(show l)++".")
getChildProcess l@(ITE _ _ _) = error ("Cannot get one child of a if-then-else node"++(show l)++".")
getChildProcess (Hiding _ pr) = pr
getChildProcess (PRecur _ _ _ _ _ pr) =  pr
getChildProcess l@(PRecCall _ _ _) = error ("Not child in a Recursive Call node"++(show l)++".")
getChildProcess PEnd =  error "Not child in a End node" 





{- ------------------------------------------------------------------- -}
{-                           PARTICIPANTS                              -}
{- ------------------------------------------------------------------- -}


--
-- Returns all the participant of a global specification
--
allParticipants :: DistributedInteraction -> Set Participant
allParticipants (ParallelInter []) = Set.empty
allParticipants (ParallelInter ((Conc x):xs)) = (allParticipants x) `Set.union`
                                                (allParticipants (ParallelInter xs))
allParticipants (SingleInter int) =
    case int of (ValPass p1 p2 _ _ _ dint) -> if (p1 /= p2) 
                                              then Set.insert p1 (Set.insert p2 (allParticipants dint))
                                              else error "Self-communication"
                (Deleg p1 p2 _ _ _ dint) -> if (p1 /= p2)
                                         then Set.insert p1 (Set.insert p2 (allParticipants dint))
                                         else error "Self-communication"
                (LabelPass p1 p2 _ _ brs) -> if (p1 /= p2) 
                                          then Set.insert p1 (Set.insert p2 (allParticipants_Branch brs))
                                          else error "Self-communication"
                Recur v es vs a dint ->  allParticipants dint  
                otherwise -> Set.empty   
                                           
allParticipants_Branch :: [BranchElmt] -> Set Participant    
allParticipants_Branch [] = Set.empty              
allParticipants_Branch ((BranchingElmt assert l x):xs) = (allParticipants x) `Set.union` 
                                                         (allParticipants_Branch xs)

{- ------------------------------------------------------------------- -}
{-                           EXPRESSIONS                               -}
{- ------------------------------------------------------------------- -}


--
-- Returns free variable 
--
variableSet :: Exp -> Set Variable
variableSet (ExpV v) = Set.singleton v
variableSet (BNot e) = variableSet e
variableSet (EBin e1 b e2) = Set.union (variableSet e1) (variableSet e2)
variableSet (EAr e1 b e2) = Set.union (variableSet e1) (variableSet e2)
variableSet (EExist v e) = (variableSet e) `Set.difference` (Set.fromList (List.map (\(VarDeclaration vr t) -> vr) v))
variableSet (EForall v e) = (variableSet e) `Set.difference` (Set.fromList (List.map (\(VarDeclaration vr t) -> vr) v))
variableSet e = Set.empty

--
-- Free variable and their type from an expression
--
variableMap :: Sorting -> Exp -> Map Variable Typ
variableMap sort (ExpV v) = case (lookupVariable sort v) of 
                              Just x -> M.singleton v x
                              Nothing -> error $ "Variable "++(show v)++" not found."
variableMap sort (BNot e) = variableMap sort e
variableMap sort (EBin e1 b e2) = M.union (variableMap sort e1) (variableMap sort e2)
variableMap sort (EAr e1 b e2) = M.union (variableMap sort e1) (variableMap sort e2)
variableMap sort (EExist v e) = (variableMap (addVariables sort v) e) 
                                `M.difference` (M.fromList (List.map (\(VarDeclaration vr t) -> (vr,t)) v))
variableMap sort (EForall v e) = (variableMap (addVariables sort v) e) 
                                 `M.difference` (M.fromList (List.map (\(VarDeclaration vr t) -> (vr,t)) v)) 
variableMap sort e = M.empty

--
-- Returns the set of variable from a list of declaration
--
variable :: [VarDecl] -> Set Variable
variable [] = Set.empty
variable ((VarDeclaration v t):vs) = Set.union (Set.singleton v) (variable vs) 

--
-- Returns the set of variable from a list of (declaration, location)
--
variableloc :: [VarDeclLoc] -> Set Variable
variableloc v = variable $ List.map (\(VarDeclLoc vd loc) -> vd) v 

--
-- Apply a substitution to an expression
--
substitute ::  Substitution -> Exp -> Exp
substitute (es, vs) (ExpV v) = case elemIndex v vs of
                              Nothing -> (ExpV v) -- variable not in the list so, it stays.
                              Just i -> es !! i
substitute (es, vs) (BNot e) = BNot (substitute (es, vs) e)
substitute (es, vs) (EBin e1 b e2) = EBin (substitute (es, vs) e1) b (substitute (es, vs) e2)
substitute (es, vs) (EAr e1 b e2) = EAr (substitute (es, vs) e1) b (substitute (es, vs) e2)
substitute (es, vs) (EExist v e) = EExist v (substitute (es, vs) e)
substitute (es, vs) (EForall v e) = EForall v (substitute (es, vs) e)
substitute (es, vs) e = e

--
-- Update a substitution
--
updateSubst :: Substitution -> [Exp] -> [VarDecl] -> Substitution
updateSubst (es, vs) exps decls = ((es++nexps), vs++ (List.map (\(VarDeclaration v t) -> v) decls))
    where nexps = List.map (\x -> substitute (es,vs) x) exps

-- TODO: ARE THESE TWO USEFUL AT ALL? 

--
-- Simplify an expression
-- (avoid True && True etc.)
--
simplify :: Assertion -> Assertion
simplify (Assert a) = Assert $ simplify_exp a

simplify_exp :: Exp -> Exp
simplify_exp (EBin BTrue  EOr   _     ) = BTrue
simplify_exp (EBin _      EOr   BTrue ) = BTrue
simplify_exp (EBin BFalse EOr   e     ) = simplify_exp e
simplify_exp (EBin e      EOr   BFalse) = simplify_exp e

simplify_exp (EBin e      EAnd  BTrue ) = simplify_exp e
simplify_exp (EBin e      EAnd  BFalse) = BFalse
simplify_exp (EBin BTrue  EAnd  e     ) = simplify_exp e
simplify_exp (EBin BFalse EAnd  e     ) = BFalse

simplify_exp (EBin _      EImpl BTrue ) = BTrue
simplify_exp (EBin BFalse EImpl _     ) = BTrue
simplify_exp (EBin BTrue  EImpl e     ) = e
simplify_exp (EBin e      EImpl BFalse) = BNot e

simplify_exp (EExist  v BTrue) = BTrue
simplify_exp (EExist  v e    ) = EExist  v $ simplify_exp e

simplify_exp (EForall v BTrue) = BTrue
simplify_exp (EForall v e    ) = EForall v $ simplify_exp e

simplify_exp (BNot      e    ) = BNot      $ simplify_exp e
simplify_exp exp = exp



--
-- This is used to avoid most of (True && True ...) in projection and types
--
mkAnd :: Exp -> Exp -> Exp
mkAnd e1 e2 = mkExp e1 EAnd e2

mkExp :: Exp -> BinOp -> Exp -> Exp
mkExp BTrue  EOr  e  = BTrue
mkExp BFalse EOr  e  = e
mkExp BTrue  EAnd e  = e
mkExp BFalse EAnd e  = BFalse
mkExp e1    op    BTrue = mkExp  BTrue  op e1
mkExp e1    op    BFalse = mkExp BFalse op e1
mkExp e1    op    e2     = EBin e1 op e2

{- ------------------------------------------------------------------- -}
{-                           ERROR MESSAGES                            -}
{- ------------------------------------------------------------------- -}

--
-- When one participant appears in different thread
--
errorConcurrency :: Participant -> a
errorConcurrency (Parti (Ident p)) = error ("[Projection] Participant "++
                                            (show p) ++
                                            " appears in several concurrent threads.")
--
-- When a participant behaves differently in different branches
-- (and is not involved in the branching)
--
errorBranch :: Participant -> LocalType -> LocalType -> a
errorBranch (Parti (Ident p)) lt1 lt2 = error ("Participant "++
                                               (show p)++
                                               " does not have the same behaviour in different branches: \n *:"++
                                               (prPrint lt1)++"\n *:"++(prPrint lt2))



--
-- Error when a lack of dependency is detected between two prefixes
-- 
errorGT :: String -> Prefix -> Prefix -> a
errorGT s (Prefix p1 p k1) (Prefix p2 q k2) =
    error ("Linearity: "++ s ++ " dependency missing " ++ " between " 
           ++ show p1 ++ "->" ++ show p ++ ":" ++ show k1
           ++ " and "
           ++ show p2 ++ "->" ++ show q ++ ":" ++ show k2)
    where show (Ident x) = x




--
-- Replaces recursive call to "t" by "end"
--
addEnd :: Variable -> Process -> Process
addEnd v p = case p of
    SingleProcess pr -> SingleProcess (f pr)
    ParallelProcess ps -> ParallelProcess (List.map (\(ConcP x) -> ConcP (addEnd v x)) ps)
  where f (PInit c ps cs pr') = (PInit c ps cs (addEnd v pr'))
        f (PJoin ch p cs pr') = (PJoin ch p cs (addEnd v pr'))
        f (PBranch c id xs) = (PBranch c id (List.map (\(PBranchElmt a l pr') -> (PBranchElmt a l (addEnd v pr'))) xs))
        f (PSelect c a id l pr) = (PSelect c a id l (addEnd v pr))
        f (ITE exp pr1 pr2) = (ITE exp (addEnd v pr1) (addEnd v pr2))
        f (PRecCall v' es cs) = if v == v'
                                then PEnd
                                else (PRecCall v' es cs)
        f (PRecur v' es cs vs cs' pr') = f (PRecur v' es cs vs cs' (addEnd v pr'))
        f (PDelegS c cs dl a pr') = (PDelegS c cs dl a (addEnd v pr'))
        f (PDelegR c dl a pr') = (PDelegR c dl a (addEnd v pr'))
        f (PSend c es vd a pr') = (PSend c es vd a (addEnd v pr')) 
        f (PReceive c dl a pr') = (PReceive c dl a (addEnd v pr'))
        f PEnd = PEnd
                 
