-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module checks that a global type is linear

module Global.Linearity
    (linearityChecker)
    where

import Common.Absglobaltype
import Common.Utils
import Global.OneTimeUnfolding
import Common.PrettyPrinter





--
-- Returns True if the DistributedInteractions are linear
--
linearityChecker :: DistributedInteraction -> Bool
linearityChecker x = (checkLin (unfoldOneTime x) [])                                



{- ------------------------------------------------------------------- -}
{-              OUTPUT / INPUT DEPENDENCY DEFINITION                   -}
{- ------------------------------------------------------------------- -}

--           
-- Input Dependency between two prefixes
-- Prefix1 < Prefix2
--
dep_ii :: Prefix -> Prefix -> Bool
dep_ii (Prefix p1 p k1) (Prefix p2 q k2)
    | k1 /= k2 = (q == p)
dep_ii (Prefix p1 p k1) (Prefix p2 q k2)
    | (k1 == k2) = (p1 == p2) && (p == q) 
dep_ii _ _ = False

--
-- InputOutput Dependency between two prefixes
-- Prefix1 < Prefix2
--
dep_io :: Prefix -> Prefix -> Bool
dep_io (Prefix p1 p k1) (Prefix q p2 k2)
    | k1 /= k2 = (q == p)
dep_io _ _ = False

--
-- Output Dependency between two prefixes
-- Prefix1 < Prefix2
dep_oo :: Prefix -> Prefix -> Bool
dep_oo (Prefix p p1 k1) (Prefix q p2 k2)
    | k1 == k2 = (q == p) 
dep_oo _ _ = False

--
-- True if there is an input dependency from the first element of the list to the last one, false otherwise
--
input_chain :: [Prefix] -> Bool
input_chain [] = error "[input_chain]empty prefix list"
input_chain (x:[]) = error "[input_chain]only one prefix in the list"
input_chain (x:y:[]) = if (dep_ii x y)  || (dep_io x y)
                       then True
                       else errorGT "II or IO" x y
input_chain (x:y:xs) = if (dep_ii x y) || (dep_io x y)
                       then input_chain (y:xs)
                       else input_chain (x:xs)

--
-- True if there is an output dependency from the first element of the list to the last one, false otherwise
--
output_chain :: [Prefix] -> Bool
output_chain [] = error "[output_chain] empty prefix list"
output_chain (x:[]) = error "[output_chain] only one prefix in the list"
output_chain (x:y:[]) = if (dep_io x y) || (dep_oo x y) 
                        then True
                        else if (dep_oo x y)
                             then errorGT "IO" x y
                             else errorGT "OO or IO" x y
output_chain (x:y:xs) = if (dep_io x y) || (dep_oo x y) 
                          then output_chain (y:xs)
                          else output_chain (x:xs)



{- ------------------------------------------------------------------- -}
{-               LINEARITY ON A SPECIFIED CHANNEL                      -}
{- ------------------------------------------------------------------- -}

--
-- True if the two prefixes may
-- induce a conflict
--
isRisk :: Prefix -> Prefix -> Bool
isRisk (Prefix p1 p2 ch) (Prefix q1 q2 k) = (k == ch) 
                                            -- && ((p1 /= q1) || (p2 /= q2))
                                            -- && ((p1 /= q2) && (p2 /= q1))

--
-- Checks whether or not there is a conflict on the channel contained in "Prefix"
-- [Prefix] contains the list of prefix encountered so far and still to be checked
-- for input/output dependency
--
channelCheck :: DistributedInteraction -> Prefix -> [Prefix] -> Bool
channelCheck (ParallelInter []) _ _ = True
channelCheck (ParallelInter ((Conc x):xs)) pre preList = (channelCheck x pre preList)
                                                         && channelCheck (ParallelInter xs) pre preList
                                                         && noChannelRace ((Conc x):xs)
channelCheck (SingleInter x) pre preList = channelCheck_I x pre preList




--
-- Returns True if no common channel is used in two different concurrent branches
--
noChannelRace :: [ConcurInteraction] -> Bool
noChannelRace [] = True
noChannelRace xs = intersects (Prelude.map (channelList . extract) xs)
    where extract (Conc y) = y 



-- Checks whether or not there is a conflict on the channel contained in "Prefix"
-- [Prefix] contains the list of prefix encountered so far and still to be checked
-- for input/output dependency
--
channelCheck_I :: Interaction -> Prefix -> [Prefix] -> Bool
channelCheck_I (ValPass (Parti p1) (Parti p2) (Chan k) _ _ int) pre preList =
    if isRisk (Prefix p1 p2 k) pre
    then (input_chain newPreList) && (output_chain newPreList)
             && (channelCheck int pre [Prefix p1 p2 k])
    else channelCheck int pre newPreList
    where newPreList = preList ++ [Prefix p1 p2 k] 

channelCheck_I (Deleg (Parti p1) (Parti p2) (Chan k) _ _ int) pre preList =
    if isRisk (Prefix p1 p2 k) pre
    then (input_chain newPreList) && (output_chain newPreList)
             && (channelCheck int pre [Prefix p1 p2 k])
    else channelCheck int pre newPreList
    where newPreList = preList ++ [Prefix p1 p2 k]

channelCheck_I (LabelPass (Parti p1) (Parti p2) (Chan k) _ xs) pre preList =
    if isRisk (Prefix p1 p2 k) pre
    then (input_chain newPreList) && (output_chain newPreList)
             && (channelCheck_B xs pre [Prefix p1 p2 k])
    else  (channelCheck_B xs pre newPreList)
    where newPreList = preList ++ [Prefix p1 p2 k]
channelCheck_I End _ preList =  True
channelCheck_I (Recur v es vs a d) p ps = channelCheck d p ps
channelCheck_I p _ _ = error ((show p)++" unsupported in Linearity check.")


--
-- Checks whether or not there is a conflict on the channel contained in "Prefix"
-- [Prefix] contains the list of prefix encountered so far and still to be checked
-- for input/output dependency
--
channelCheck_B :: [BranchElmt] -> Prefix -> [Prefix] -> Bool
channelCheck_B [] _ _ = True
channelCheck_B (x:xs) pre preList = 
    (channelCheck_E x pre preList)
    && (channelCheck_B xs pre preList)

-- channelCheck_B helper
channelCheck_E :: BranchElmt -> Prefix -> [Prefix] -> Bool
channelCheck_E (BranchingElmt _ _ int) pre preList = channelCheck int pre preList



{- ------------------------------------------------------------------- -}
{-                     LINEARITY CHECK ON EVERY CHANNEL                -}
{- ------------------------------------------------------------------- -}

--
-- Returns True if there is a prefix (in the list [Prefix])
-- such that its channel is the same as the one specified (Ident) 
--
isIn :: Ident -> [Prefix] -> Bool
isIn _ [] = False
isIn ch ((Prefix p1 p2 k):xs) = (ch == k) || (isIn ch xs)

--
-- Returns True if the DistributedInteractions are linear
-- the list of Prefix contains the prefixes for which
-- linearity check has been done
-- 
checkLin :: DistributedInteraction -> [Prefix] -> Bool
checkLin (ParallelInter []) _ = True
checkLin (ParallelInter ((Conc x):xs)) pre = (checkLin x pre) && (checkLin (ParallelInter xs) pre)
checkLin (SingleInter x) pre = checkLin_I x pre
                               
--
-- checkLin helper
--
checkLin_I :: Interaction -> [Prefix] -> Bool
checkLin_I (ValPass (Parti p1) (Parti p2) (Chan k) _ _ int) preList =
    if (k `isIn` preList)
    then (checkLin int preList)
    else (channelCheck int (Prefix p1 p2 k) [Prefix p1 p2 k]) && (checkLin int newPreList)
 where newPreList =  preList ++ [Prefix p1 p2 k] 

checkLin_I (Deleg (Parti p1) (Parti p2) (Chan k) _ _ int) preList =
    if (k `isIn` preList)
    then (checkLin int preList)
    else (channelCheck int (Prefix p1 p2 k) [Prefix p1 p2 k]) && (checkLin int newPreList)
    where newPreList =  preList ++ [Prefix p1 p2 k] 

checkLin_I (LabelPass (Parti p1) (Parti p2) (Chan k) id ((BranchingElmt _ _ int):xs)) preList =
    if (k `isIn` preList)
    then (checkLin int preList) 
             && (checkLin_I (LabelPass (Parti p1) (Parti p2) (Chan k) id xs) preList)
    else (channelCheck int (Prefix p1 p2 k) [Prefix p1 p2 k])
             && (checkLin int newPreList)
             && checkLin_I (LabelPass (Parti p1) (Parti p2) (Chan k) id xs) preList
    where newPreList =  preList ++ [Prefix p1 p2 k] 
checkLin_I (LabelPass (Parti p1) (Parti p2) (Chan k) id []) preList = True
checkLin_I End _ = True
checkLin_I (Recur v es vs a d) ps = checkLin d ps
checkLin_I p pre = error $ (show p)++" unsupported in Linearity check ("++(show pre)++")."
