-- Author: Julien Lange (jl250@le.ac.uk)
-- 2009-2010

-- This module implements the core of the typing system


module Local.TypingSystem (sessionTypingConc) where

import Common.Typing
import Common.Absglobaltype
import Common.Utils
import Data.Map  as M 
import Data.List as L
import Data.Set as S
import Common.PrettyPrinter
import Common.ExpressionTyping
import Global.WellAsserted
import Assertion.LogicAPI
import Data.Maybe

-- Function used to test compability of two types
type Compatible  = Typing -> Typing -> Bool

-- Function used to compose two types
type Composition = Typing -> Typing -> Typing

-- Type of a typing system
-- (e.g. binarySessionTyping and typingSystem)
type TypingSys   = Sorting -> Process -> Typing


-- type Compare     = LocalType -> LocalType -> Bool



--
-- Types a process according to an environment and the two functions for compability and composition
-- (Top level typing function)
--
sessionTypingConc :: Compatible -> Composition -> Sorting -> Process -> Typing
sessionTypingConc compatible compose sort (SingleProcess pr) =
    sessionTyping (sessionTypingConc compatible compose) sort emptySub BTrue pr
sessionTypingConc compatible compose sort (ParallelProcess ps) =
    sessionTypingConc_helper compatible compose sort ps

--
-- Helper for concurrent branches
-- (here the compability check and composition are done)
--
sessionTypingConc_helper :: Compatible -> Composition -> Sorting -> [ConcurProcess] -> Typing
sessionTypingConc_helper compatible compose sort (ConcP p:xs) =
    if compatible typep rest
    then compose typep rest
    else error "Concurrent processes' types are not compatible"
        where typep = sessionTypingConc compatible compose sort p
              rest = sessionTypingConc_helper compatible compose sort xs
sessionTypingConc_helper compatible compose sort [] = []



--
-- Starts a type every type a session initiation primitive is found
-- in case projections are available the type is matched against it
--
sessionTyping :: TypingSys -> Sorting -> Substitution -> Exp -> Primitive -> Typing
sessionTyping ftype sort sub aenvi (PInit ch (p:ps) cs pr) = 
    case sproj of
      Nothing -> styping
      Just x -> if (refinement stype x) -- type has to refine the projection
                then styping
                else error ("LocalType doesn't match projection for "++(prPrint p)++"!\nType:\n"
                         ++(prPrint stype)++"\n\nProjection:\n"
                         ++(prPrint x)++".")
    where stype = simpleTyping sort cs sub aenvi pr
          sproj = lookupLocalType sort ch p
          styping = (ch, p, cs, stype):(ftype sort pr)
sessionTyping ftype sort sub aenvi (PJoin ch p cs pr) =
    case sproj of
      Nothing -> styping
      Just x -> if (refinement stype x)
                then styping
                else error ("LocalType doesn't match projection for "++(prPrint p)++"!\nType:\n"
                         ++(prPrint stype)++"\n\nProjection:\n "
                         ++(prPrint x)++".")
    where stype = simpleTyping sort cs sub aenvi pr
          sproj = lookupLocalType sort ch p
          styping = (ch, p, cs, stype):(ftype sort pr)
sessionTyping ftype sort sub aenvi (PBranch c id ((PBranchElmt ass l pr):xs)) = 
    (ftype sort pr) ++ (sessionTyping ftype sort sub aenvi (PBranch c id xs)) -- should check that a same session is not typed twice 
sessionTyping ftype sort sub aenvi (ITE exp pr1 pr2) =
    if (expTyping sort exp) == Just TBool &&  (tp1 == tp2)
    then tp1
    else error "typing problems in if-then-else"
    where tp1 = ftype sort pr1
          tp2 = ftype sort pr2
sessionTyping _ sort sub aenvi (PRecCall _ _ _) = [] -- match with recursion stuff
sessionTyping _ sort sub aenvi PEnd = []
sessionTyping ftype sort sub aenvi (PBranch c id []) = [] 
sessionTyping ftype sort sub aenvi (PRecur v exps cs1 vs cs2 pr) = ftype (addVariables sort vs) pr
sessionTyping ftype sort sub aenvi (PSend ch exps vars (Assert assert) proc) =
    ftype (addVariables sort vars) proc
sessionTyping ftype sort sub aenvi (PReceive ch vars (Assert assert) proc) =
    ftype (addVariables sort vars) proc
sessionTyping ftype sort sub aenvi pr = ftype sort (getChildProcess pr)




--
-- Types a vector of channels (belonging to the same session)
-- Whole process
-- (sub top level function)
--
simpleTyping :: Sorting -> [Channel] -> Substitution -> Exp -> Process -> LocalType
simpleTyping sort chs sub aenvi (SingleProcess prim) = primitiveTyping sort chs sub aenvi prim
simpleTyping sort chs sub aenvi (ParallelProcess xs) = simpleTyping_helper sort chs sub aenvi xs []


simpleTyping_helper :: Sorting -> [Channel] -> Substitution -> Exp -> [ConcurProcess] -> [LocalType] -> LocalType
simpleTyping_helper sort chs sub aenvi [] acc = singleType acc LEnd
simpleTyping_helper sort chs sub aenvi ((ConcP x):xs) acc = simpleTyping_helper sort chs sub aenvi xs (typex:acc)
    where typex = simpleTyping sort chs sub aenvi x

--
-- Check for multithreading and returns the only relevant type
-- (participant must be single threaded)
--
singleType :: [LocalType] -> LocalType -> LocalType
singleType [] lt = lt 
singleType (x:xs) lt = if ((x /= LEnd) && (lt == LEnd)) || (x == LEnd)
                       then singleType xs x
                       else error "Participant is not single threaded!"


--
-- Types a vector of channels (belonging to the same session)
-- Primitives only
-- (low level function)
--
primitiveTyping :: Sorting -> [Channel] -> Substitution -> Exp -> Primitive -> LocalType
primitiveTyping sort chs sub aenvi (PInit _ _ _ pr) = simpleTyping sort chs sub aenvi pr

primitiveTyping sort chs sub aenvi (PJoin _ _ _ pr) = simpleTyping sort chs sub aenvi pr
    
primitiveTyping sort chs sub aenvi (PBranch ch id brs) =
    if (ch `elem` chs)
    then Branch ch id brty
    else branchTyping_check brty
    where brty = branchTyping sort chs sub aenvi brs

primitiveTyping sort chs sub aenvi (PSelect ch (Assert assert) id label pr) =
    if (ch `elem` chs)
    then if checkSatis sort (EBin aenvi EImpl assert)
            then Select ch id [LabelElmt (Assert assert) label (simpleTyping sort chs sub aenvi pr)]
            else error $ "[Typing-Select] Assertion not satisfiable: "++(prPrint (EBin aenvi EImpl assert))++"."
    else simpleTyping sort chs sub aenvi pr

primitiveTyping sort chs sub aenvi (PSend ch exps vars (Assert assert) proc) =
    if (ch `elem` chs)
    then if  checkSatis upsort (EBin aenvi EImpl nassert)
         then Send ch vars (Assert assert) (simpleTyping upsort chs nsub aenvi proc)
         else error $ "[Typing-Send] Assertion not satisfiable: "++(prPrint (EBin aenvi EImpl nassert))++"."
    else simpleTyping upsort chs nsub aenvi proc
    where expT = expsTyping upsort exps -- type of exps
          upsort = addVariables sort vars
          nexps = L.map (substitute sub) exps
          nsub = updateSubst sub exps vars
          nassert = substitute nsub assert

primitiveTyping sort chs sub aenvi (PDelegS ch bchs del@(SessPass chs' lt p) ass proc) =
-- should check that bchs are not used anymore!
     if (ch `elem` chs)
     then LDelegS ch del ass (simpleTyping sort chs sub aenvi proc)
     else if chs' == chs 
          then lt
          else simpleTyping sort chs sub aenvi proc
-- if chs' == chs

primitiveTyping sort chs sub aenvi (PDelegR ch del@(SessPass chs' lt p) ass proc) =
    if (ch `elem` chs)
    then if (g == lt)
         then LDelegR ch del ass (simpleTyping sort chs sub aenvi proc)
         else error $ "Channels "++(prPrintList chs')
                  ++" not used as advertised: \n*[Inferred  ]"
                  ++(prPrint g)++"\n*[Advertised]"++(prPrint lt)++"."
    else simpleTyping sort chs sub aenvi proc
    where g = simpleTyping sort chs' sub aenvi proc

primitiveTyping sort chs sub aenvi (PReceive ch vars (Assert assert) proc) =
    if (ch `elem` chs)
    then Receive ch vars (Assert assert) (simpleTyping  upsort chs sub naenvi proc)
    else simpleTyping upsort chs sub aenvi proc
    where naenvi = (mkAnd aenvi assert)
          upsort = addVariables sort vars

primitiveTyping sort chs sub aenvi (ITE exp pr1 pr2) =
    if (expTyping sort exp) == Just TBool
    then mergeSelect tp1 tp2
    else error ("Expression is not boolean in an if-then-else (on channels "++(prPrintList chs)++").")
    where tp1 = simpleTyping sort chs sub (EBin aenvi EAnd exp) pr1
          tp2 = simpleTyping sort chs sub (EBin aenvi EAnd (BNot exp)) pr2

primitiveTyping sort chs sub aenvi (PRecur v exs cs1 vs cs2 pr) =  (LRecur v exs vs (Assert BTrue) proj)
    where proj = if length interc == 0 
                   then LEnd -- (channels are not used in the recursion)--  simpleTyping nsort chs nsub naenvi pr
                   else renameChannels (simpleTyping nsort ncs2 nsub naenvi pr) ncs2 interc
          nsort = addVariables nsort' vs
          nsort' = addProcess sort v (vs, cs2, addEnd v pr)
          interc = intersect cs1 chs
          idx = findIndices (\x -> x `elem` interc) cs1
          ncs2 = L.map (\x -> cs2 !! x) idx
          nsub = updateSubst sub exs vs
          naenvi = substitute nsub aenvi
          
primitiveTyping sort chs sub aenvi  (PRecCall v exs cs) = 
    if (catMaybes (expsTyping sort exs)) == (L.map (\(VarDeclaration v t ) -> t ) vt) && cs == cs'
        && (primitiveTyping sort chs nsub aenvi (PRecur v exs cs [] cs' pr)) /= (LEnd)
    then LRecCall v exs
    else error $ "Recursive call parameters do not match recursive definition."
      where (vt , cs', pr) = lookupProcess sort v                
            nsub = updateSubst sub exs vt
                 
primitiveTyping _ _ _ _ PEnd = LEnd

-- primitiveTyping s c p = error $ (show s) ++ "\n" ++ (show c) ++ "\n" ++  (show p)


--
-- Types branches
--
branchTyping :: Sorting -> [Channel] -> Substitution -> Exp -> [PBranchElmt] -> [LabelElmt]
branchTyping _ _ _ _[] = []
branchTyping sort chs sub aenvi ((PBranchElmt (Assert ass) l pr):xs) = 
    (LabelElmt (Assert ass) l (simpleTyping sort chs sub (mkAnd aenvi ass) pr)):(branchTyping sort chs sub aenvi xs)

branchTyping_check :: [LabelElmt] -> LocalType
branchTyping_check [] = error "Branching is empty"
branchTyping_check [(LabelElmt _ _ lt)] = lt
branchTyping_check (e1@(LabelElmt a1 l1 lt1):e2@(LabelElmt a2 l2 lt2):xs) = 
    mergeSelect lt1 (branchTyping_check (e2:xs))

--
-- Merges types to gather Select primitive belonging to the same branching construct
--
mergeSelect :: LocalType -> LocalType -> LocalType
mergeSelect (Send c1 t1 a1 lt1)       (Send c2 t2 a2 lt2)
    | (c1 == c2) && (t1 == t2)                               = Send c1 t1 a1 (mergeSelect lt1 lt2)
mergeSelect (Receive c1 t1 a1 lt1)    (Receive c2 t2 a2 lt2)
    | (c1 == c2) && (t1 == t2)                               = Receive c1 t1 a1 (mergeSelect lt1 lt2)
mergeSelect (Branch c1 id1 ls1)       (Branch c2 id2 ls2)
    | (c1 == c2) && (id1 == id2) && (ls1 == ls2)             = Branch c1 id1 (mergeSelect_Branch ls1 ls2)
mergeSelect (Select c1 id1 xs)        (Select c2 id2 ys) 
    | (c1 == c2) && (id1 == id2)                             = Select c1 id1 (xs++ys) -- Select merging
mergeSelect (LRecur v1 e1 vd1 a1 lt1) (LRecur v2 e2 vd2 a2 lt2)
    | (v1 == v2) && (e1 == e2) && (vd1 == vd2) && (a1 == a2) = LRecur v1 e1 vd1 a1 (mergeSelect lt1 lt2)
mergeSelect (LRecCall v1 e1)          (LRecCall v2 e2)
    | (v1 == v2) && (e1 == e2)                               = LRecCall v1 e1
mergeSelect LEnd                      LEnd                   = LEnd
mergeSelect Bottom                    Bottom                 = Bottom
mergeSelect lt1                       (LRecCall _ _)         = lt1 -- if the type is only a call to recursion, we "skip" it.
mergeSelect (LRecCall _ _)            lt1                    = lt1
mergeSelect lt1 lt2 = error ("[mergeSelect]Type "++(prPrint lt1)++" and type "++(prPrint lt2)++" are incompatible")

--
-- Helpers for mergeSelect
-- FIXME: /!\ We here assume that labels are always used in the same order
--
mergeSelect_Branch :: [LabelElmt] -> [LabelElmt] -> [LabelElmt]
mergeSelect_Branch (x:xs) (y:ys) =
    (mergeSelect_Label x y):(mergeSelect_Branch xs ys)

mergeSelect_Label :: LabelElmt -> LabelElmt -> LabelElmt
mergeSelect_Label (LabelElmt a1 l1 lt1) (LabelElmt a2 l2 lt2)
    | (a1 == a2) && (l1 == l2) = LabelElmt a1 l1 (mergeSelect lt1 lt2)
mergeSelect_Label l1 l2 = error ("Branch "
                                  ++(prPrint l1)++" and "
                                  ++(prPrint l2)++" are not equivalent.")




--
-- TODO: Refinement should give more information on why it fails!
-- (i.e. which assertions are failing)
--
-- Returns true lt1 refines lt2 
-- (see definition in "A Theory of DbC..."
--
refinement :: LocalType -> LocalType -> Bool
refinement lt1 lt2 = refinement_ (Empty, lt1) (Empty, lt2)

refinement_ :: (Sorting, LocalType) -> (Sorting, LocalType) -> Bool
refinement_ (s1, (Send c1 t1 a1 lt1))        (s2, (Send c2 t2 a2 lt2)) = 
    (c1 == c2) && (t1 == t2) && (imply (ns1, a1) (ns2, a2)) && (refinement_ (ns1, lt1) (ns2, lt2))
        where ns1 = addVariables s1 t1
              ns2 = addVariables s2 t2

refinement_ (s1, (LDelegS c1 d1@(SessPass cs1' lt1' p1) a1 lt1)) (s2, (LDelegS c2 d2@(SessPass cs2' lt2' p2) a2 lt2)) =
    (c1 == c2) && (cs1' == cs2') && (p1 == p2) && (imply (s1, a1) (s2, a2)) && (refinement_ (s1, lt1') (s2, lt2')) && (refinement_ (s1, lt1) (s2, lt2))

refinement_ (s1, (Receive c1 t1 a1 lt1))     (s2, (Receive c2 t2 a2 lt2)) =
    (c1 == c2) && (t1 == t2) && (imply (ns2, a2) (ns1, a1)) && (refinement_ (ns1, lt1) (ns2, lt2))
        where ns1 = addVariables s1 t1
              ns2 = addVariables s2 t2

refinement_ (s1, (LDelegR c1 d1@(SessPass cs1' lt1' p1) a1 lt1)) (s2, (LDelegR c2 d2@(SessPass cs2' lt2' p2) a2 lt2)) =
    (c1 == c2) && (cs1' == cs2') && (p1 == p2) && (imply (s2, a2) (s1, a1)) && (refinement_ (s1, lt2') (s2, lt2')) && (refinement_ (s1, lt1) (s2, lt2))

refinement_ (s1, (Branch c1 id1 ls1))        (s2, (Branch c2 id2 ls2)) = 
    (c1 == c2) && (id1 == id2) && (isSubsetOf (lbs ls2) (lbs ls1))
                   && refLabel (s2, ls2) (s1, ls1)

refinement_ (s1, (Select c1 id1 ls1))        (s2, (Select c2 id2 ls2)) =
    (c1 == c2) && (id1 == id2) && (isSubsetOf (lbs ls1) (lbs ls2))
                   && refLabel (s1, ls1) (s2, ls2)

refinement_ (s1, (LRecur v1 e1 vd1 a1 lt1))  (s2, (LRecur v2 e2 vd2 a2 lt2)) =
    -- NB: assertion are not checked as there is no assertion in
    --     the process language
    (v1 == v2) && (e1 == e2) && (vd1 == vd2) && (refinement_ (ns1, lt1) (ns2, lt2))
        where ns1 = addVariables s1 vd1
              ns2 = addVariables s2 vd2
refinement_ (s1, (LRecCall v1 e1))           (s2, (LRecCall v2 e2)) =
    (v1 == v2) && (e1 == e2)
refinement_ (s1, LEnd)                     (s2, LEnd)                        = True
refinement_ (s1, Bottom)                   (s2, Bottom)                      = True
refinement_ (s1, LEnd)                     (s2, LRecur _ _ _ _ LEnd)         = True
refinement_ (s1, LRecur _ _ _ _ LEnd)      (s2, LEnd)                        = True
refinement_ (s1, t1)                       (s2, t2)                          =
    error $ "[Refinement]Two types are incompatible: \n*"++(prPrint t1)++"\n*"++(prPrint t2)
              ++"\n\n"++(show s1) ++"\n\n"++(show s2)
    

-- Check that e1 => e2
imply :: (Sorting, Assertion) -> (Sorting, Assertion) -> Bool
imply (Empty, (Assert e1)) (Empty, (Assert e2)) =
    if sat
    then True
    else error $ "[Refinement] assertion ["++(prPrint e1)++"] does not refine ["++(prPrint e2)++"]."
    where sat = checkSatis Empty (EBin e1 EImpl e2)
imply ((Sorting (NameMap s1) pm prm ), (Assert e1)) ((Sorting (NameMap s2) pm' prm' ), (Assert e2)) =
    if sat
    then True
    else error $ "[Refinement] assertion ["++(prPrint e1)++"] does not refine ["++(prPrint e2)++"]."
    where sat = checkSatis sort (EBin e1 EImpl e2)
          sort = (Sorting (NameMap (M.union s1 s2)) emptyProcessMap emptyProjMap) -- ProcMap and ProjMap are empty anyways


-- Refinement helpers:
--
-- Checks that a list of "LabelElmt" (ls1) refines another (ls2) 
-- i.e. Ai => Bi && refinement Ti Tj
-- 
-- Basically, we sort the labels in each list, then we compare their child processes
-- Two by two
--

--refLabel (s1, ls1) (s2, ls2) = foldl (&&) True (L.map (\(x,y) -> x && y)
--                                        (mergeAscList (\x y -> checkSatis Empty ((EBin x EImpl y))) (\z w -> refinement_ (s1, w) (s2, z)) (asss ls1) (asss ls2)))

refLabel (s1, ls1) (s2, ls2) = 
    foldl (&&) True (L.map (\(x,y) -> x && y)
                          (
                           mergeAscList (\x y -> imply (s1, (Assert x)) (s2, (Assert y)))
                                            (\z w -> refinement_ (s1, w) (s2, z)) (asss ls1) (asss ls2)
                          )
                    )
-- 
-- NB: We assume: {k2} included in {k1}
--
mergeAscList :: (Eq k, Show k) => (a -> a -> b) -> (lt -> lt -> c)-> [(k,(a,lt))] -> [(k,(a,lt))] -> [(b,c)]
mergeAscList _ _ [] _ = []
mergeAscList _ _ _ [] = []
mergeAscList f g ((k,(a,t)):xs) ((k',(a',t')):ys) = if k == k'
                                                    then ((f a a'), (g t t')):(mergeAscList f g xs ys)
                                                    else error $ "[Refinement] List are not sorted or have different keys ("++(show k')++")."
                
-- make set of labels
lbs ls = S.fromList (L.map (\(LabelElmt a la lt) -> la) ls)
-- generate list of tuple: (k,(a,t)) == (label,(assertion,localtype))
asss ls =  M.toAscList $ M.fromList $ L.map (\(LabelElmt (Assert a) la lt) -> (la,(a,lt))) ls

--
--


--
-- Utils
--
declTyping :: Sorting -> [VarDecl] -> [Typ] -- type of a list of variable declaration (ex Receive)
declTyping _ [] = []
declTyping s ((VarDeclaration v t):xs) = t:(declTyping s xs)


--
-- Renames channels in a local type
-- This is used to check Recursive types (where channels are give as recursion parameters)
--
renameChannels :: LocalType -> [Channel] -> [Channel] -> LocalType   
-- cs: to be replaced 
renameChannels (Send c v a lt) cs cs' =
    case elemIndex c cs of 
      Nothing -> Send c v a (renameChannels lt cs cs')
      Just i -> Send (cs' !! i) v a (renameChannels lt cs cs')
renameChannels (Receive c v a lt) cs cs' = 
    case elemIndex c cs of 
      Nothing -> Receive c v a (renameChannels lt cs cs')
      Just i -> Receive (cs' !! i) v a (renameChannels lt cs cs')
renameChannels (Branch c id lb) cs cs' = 
    case elemIndex c cs of 
      Nothing -> Branch c id (L.map (\(LabelElmt a l lt) -> (LabelElmt a l (renameChannels lt cs cs'))) lb)
      Just i -> Branch (cs' !! i) id (L.map (\(LabelElmt a l lt) -> (LabelElmt a l (renameChannels lt cs cs'))) lb)
renameChannels (Select c id lb) cs cs' = 
    case elemIndex c cs of 
      Nothing -> Select c id (L.map (\(LabelElmt a l lt) -> (LabelElmt a l (renameChannels lt cs cs'))) lb)
      Just i -> Select (cs' !! i) id  (L.map (\(LabelElmt a l lt) -> (LabelElmt a l (renameChannels lt cs cs'))) lb)
renameChannels (LRecur v e vs a lt) cs cs' = LRecur v e vs a (renameChannels lt cs cs')
renameChannels t _ _ = t 
